from __future__ import absolute_import, division, print_function, unicode_literals

# The variable SYMMETRY_TOL controls the precision used by spglib in order
# to find symmetries. The atoms are allowed to move 1/2*SYMMETRY_TOL from
# their symmetry positions in order for spglib to still detect symmetries.
SYMMETRY_TOL = 2*0.2  # unit: angstrom

# The threshold for a system to be considered "flat". Used e.g. when
# determining if a 2D structure is purely 2-dimensional to allow extra rigid
# transformations that are improper in 3D but proper in 2D.
FLAT_DIM_THRESHOLD = 0.1

# An ordered list of Wyckoff letters
WYCKOFF_LETTERS = list("abcdefghijklmnopqrstuvwxyzA")
WYCKOFF_LETTER_POSITIONS = {letter: positions for positions, letter in enumerate(WYCKOFF_LETTERS)}

#===============================================================================
# Constants for classification
MAX_SINGLE_CELL_SIZE = 5
MAX_2D_CELL_HEIGHT = 5
# MAX_CELL_SIZE = [4, 12]
MAX_CELL_SIZE = [12]
# REL_POS_TOL = [0.25, 0.5, 0.75]  # Position tolerances. Given relative to minimum distance between two atoms in the system.
REL_POS_TOL = [0.25, 0.75]  # Position tolerances. Given relative to minimum distance between two atoms in the system.
# REL_POS_TOL = [0.25]  # Position tolerances. Given relative to minimum distance between two atoms in the system.
POS_TOL_SCALING = 0.0  # 0.0  # Per angstrom
ANGLE_TOL = 20
CLUSTER_THRESHOLD = 3.5
CRYSTALLINITY_THRESHOLD = 0.25
DELAUNAY_THRESHOLD = 1.5
BOND_THRESHOLD = 0.75
CHEM_SIMILARITY_THRESHOLD = 0.4
CELL_SIZE_TOL = 0.25
MAX_N_ATOMS = 1000
MIN_COVERAGE = 0.5



from __future__ import absolute_import, division, print_function

import spglib

import numpy as np

from collections import defaultdict
from operator import attrgetter

from matid.utils.segfault_protect import segfault_protect
from matid.data.symmetry_data import PROPER_RIGID_TRANSFORMATIONS, IMPROPER_RIGID_TRANSFORMATIONS
from matid.exceptions import CellNormalizationError, SystaxError
from matid.data.symmetry_data import SPACE_GROUP_INFO, WYCKOFF_SETS
from matid.data import constants
from matid.core.system import System
from matid.symmetry import WyckoffSet
import matid.geometry

from ase import Atoms


class SymmetryAnalyzer(object):
    """A base class for getting symmetry related properties of unit cells.
    """
    def __init__(self, system=None, symmetry_tol=None, min_2d_thickness=1):
        """
        Args:
            system(ASE.Atoms): The system to inspect.
            symmetry_tol(float): The tolerance for the symmetry detection.
            min_2d_thickness(float): The minimum thickness in angstroms for the
                conventional cell that is returned for 2D systems.
        """
        self._original_system = None
        self._analyzed_system = None
        self.min_2d_thickness = min_2d_thickness
        if symmetry_tol is None:
            self.symmetry_tol = constants.SYMMETRY_TOL
        else:
            self.symmetry_tol = symmetry_tol

        self.set_system(system)

    def set_system(self, system):
        """Sets a new system for analysis.
        """
        self.reset()
        self._original_system = system

        # Analyze whether the system is 2D or not. 2D systems will be first
        # un-symmetrized in the nonperiodic direction by adding sufficient
        # vacuum.

        # Determine if the system has three periodic directions or two.
        pbc = system.get_pbc()
        n_pbc = np.sum(pbc)

        # Regular bulk structures
        if n_pbc == 3:
            self._analyzed_system = system
        elif n_pbc == 2:
            # Get the index of the non-periodic axis
            i_pbc = np.argwhere(pbc == False)[0]

            # Before calculating the conventional system, make sure that there
            # is enough vacuum in the periodic direction to remove any
            # translational symmetries that are smaller than the basis vector
            # in the non-periodic direction.
            symmetry_broken_system = system.copy()
            thickness = max(5, 3*matid.geometry.get_thickness(symmetry_broken_system, i_pbc))
            old_cell = symmetry_broken_system.get_cell()
            old_basis = old_cell[i_pbc, :]
            old_basis_len = np.linalg.norm(old_basis)
            old_basis_norm = old_basis/old_basis_len
            new_basis = thickness*old_basis_norm
            old_cell[i_pbc, :] = new_basis
            symmetry_broken_system.set_cell(old_cell)
            self._analyzed_system = symmetry_broken_system
        else:
            raise ValueError(
                "No symmetry routines defined for system that do not have 3D or"
                " 2D periodicity."
            )

    def reset(self):
        """Used to reset all the cached values.
        """
        self._symmetry_dataset = None

        self._conventional_system = None
        self._conventional_wyckoff_letters = None
        self._conventional_equivalent_atoms = None
        self._conventional_lattice_fit = None

        self._spglib_conventional_system = None
        self._spglib_wyckoff_letters_conventional = None
        self._spglib_equivalent_atoms_conventional = None

        self._spglib_primitive_system = None
        self._spglib_wyckoff_letters_primitive = None
        self._spglib_equivalent_atoms_primitive = None
        self._spglib_primitive_to_original_mapping = None

        self._primitive_system = None
        self._primitive_wyckoff_letters = None
        self._primitive_equivalent_atoms = None

        self._best_transform = None

    def get_space_group_number(self):
        """
        Returns:
            int: The space group number.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["number"]

        return value

    def get_space_group_international_short(self):
        """
        Returns:
            str: The international space group short symbol.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["international"]

        return value

    def get_hall_symbol(self):
        """
        Returns:
            str: The Hall symbol.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["hall"]

        return value

    def get_hall_number(self):
        """
        Returns:
            int: The Hall number.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["hall_number"]

        return value

    def get_point_group(self):
        """Symbol of the crystallographic point group in the Hermann-Mauguin
        notation.

        Returns:
            str: point group symbol
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["pointgroup"]

        return value

    def get_is_chiral(self):
        """Returns a boolean value that tells if this object is chiral or not
        (achiral). A chiral object has symmetry operations that are all proper,
        i.e. their determinant is +1.

        Returns:
            bool: is the object chiral.
        """
        operations = self.get_symmetry_operations()
        rotations = operations["rotations"]
        chiral = True
        for rotation in rotations:
            determinant = np.linalg.det(rotation)
            if determinant == -1.0:
                return False

        return chiral

    def get_has_free_wyckoff_parameters(self):
        """Tells whether this system has Wyckoff positions with free variables.

        Returns:
            bool: Indicates the presence of Wyckoff positions with free variables.
        """
        space_group = self.get_space_group_number()
        wyckoff_letters = set(self.get_wyckoff_letters_original())
        wyckoff_info = WYCKOFF_SETS[space_group]
        for wyckoff_letter in wyckoff_letters:
            variables = wyckoff_info[wyckoff_letter]["variables"]
            if len(variables) != 0:
                return True
        return False

    def get_crystal_system(self):
        """Get the crystal system based on the space group number. There are
        seven different crystal systems:

            - Triclinic
            - Monoclinic
            - Orthorhombic
            - Tetragonal
            - Trigonal
            - Hexagonal
            - Cubic

        Return:
            str: The name of the crystal system.
        """
        space_group = self.get_space_group_number()
        crystal_system = SPACE_GROUP_INFO[space_group]["crystal_system"]

        return crystal_system

    def get_bravais_lattice(self):
        """Return Bravais lattice in the Pearson notation, where the first
        lowercase letter indicates the crystal system, and the second uppercase
        letter indicates the centring type.

        Crystal system letters:

            - a = triclinic
            - m = monoclinic
            - o = orthorhombic
            - t = tetragonal
            - h = hexagonal and trigonal
            - c = cubic

        Lattice type letters:

            - P = Primitive
            - S (= A or B or C) = One side/face centred
            - I = Body centered
            - R = Rhombohedral centring
            - F = All faces centred

        :param crystal_system: The crystal system
        :param space_group: The space group number.
        :type crystal_system: str
        :type space_group: int

        :return: The Bravais lattice in the Pearson notation.
        :rtype: str
        """

        space_group = self.get_space_group_number()
        if space_group is None:
            return None

        bravais_lattice = SPACE_GROUP_INFO[space_group]["bravais_lattice"]

        # The different one-sided centrings are merged into one letter
        if bravais_lattice[1] in ["A", "B", "C"]:
            bravais_lattice = bravais_lattice[0] + "S"

        return bravais_lattice

    def get_primitive_system(self):
        """Returns a primitive description for this system.

        This description uses a primitive lattice where positions of the
        atoms, and the cell basis vectors are idealized to follow the
        symmetries that were found with the given precision. This means that
        e.g. the volume, density, angles between basis vectors and basis vector
        lengths may have small deviations from the original system.

        Returns:
            ASE.Atoms: The primitive system.
        """
        if self._primitive_system is not None:
            return self._primitive_system

        conv_sys = self.get_conventional_system()
        conv_wyckoff = self.get_wyckoff_letters_conventional()
        conv_equivalent = self.get_equivalent_atoms_conventional()
        space_group_short = self.get_space_group_international_short()

        prim_sys, prim_wyckoff, prim_equivalent = self._get_primitive_system(
            conv_sys, conv_wyckoff, conv_equivalent, space_group_short)

        self._primitive_system = prim_sys
        self._primitive_wyckoff_letters = prim_wyckoff
        self._primitive_equivalent_atoms = prim_equivalent

        return self._primitive_system

    def get_conventional_system(self):
        """Used to get the conventional representation of this system.

        This description uses a conventional lattice where positions of the
        atoms, and the cell basis vectors are idealized to follow the
        symmetries that were found with the given precision. This means that
        e.g. the volume, density, angles between basis vectors and basis vector
        lengths may have small deviations from the original system.
        """
        if self._conventional_system is not None:
            return self._conventional_system

        # Determine if the system has three periodic directions or two.
        pbc = self._original_system.get_pbc()
        n_pbc = np.sum(pbc)

        # Regular bulk structures
        if n_pbc == 3:

            spglib_conv_sys = self._get_spglib_conventional_system()

            # Find a proper rigid transformation that produces the best combination
            # of atomic species in the Wyckoff positions.
            space_group = self.get_space_group_number()
            wyckoff_letters = self._get_spglib_wyckoff_letters_conventional()
            equivalent_atoms = self._get_spglib_equivalent_atoms_conventional()
            ideal_sys, ideal_wyckoff = self._find_wyckoff_ground_state(
                space_group,
                wyckoff_letters,
                spglib_conv_sys
            )
            ideal_sys = System.from_atoms(ideal_sys)
            ideal_sys.set_equivalent_atoms(equivalent_atoms)
            ideal_sys.set_wyckoff_letters(ideal_wyckoff)

            self._conventional_system = ideal_sys
            self._conventional_wyckoff_letters = ideal_wyckoff
            self._conventional_equivalent_atoms = equivalent_atoms
            return ideal_sys
        # 2D materials get a special treatment
        elif n_pbc == 2:

            i_pbc = np.argwhere(pbc == False)[0]

            # Get the full 3D conventional system and it's symmetries. It will
            # include some symmetries that have a translational component
            # corresponding to the non-periodic axis, but it does not matter in
            # this case.
            spglib_conv_sys = self._get_spglib_conventional_system()

            # Determine if the structure is flat. This will affect the
            # transformation that are allowed when finding the Wyckoff positions
            is_flat = False
            thickness = matid.geometry.get_thickness(self._original_system, i_pbc)
            if thickness < 0.5*self.symmetry_tol:
                is_flat = True

            # Determine the new non-periodic direction in the normalized cell.
            # The index of the originally non-periodic dimension may not correspond
            # to the one in the normalized system, because the normalized system
            # may use a different coordinate system.
            transformation_matrix = self.get_symmetry_dataset()["transformation_matrix"]
            nonperiodic_axis = None
            prec = 1e-8
            for i_axis, axis in enumerate(transformation_matrix):
                if abs(axis[i_pbc]) > prec and \
                   abs(axis[(i_pbc+1) % 3]) < prec and \
                   abs(axis[(i_pbc+2) % 3]) < prec:
                    nonperiodic_axis = i_axis
                    break
            if nonperiodic_axis is None:
                raise SystaxError(
                    "Could not detect the non-periodic direction in the normalized "
                    "2D cell."
                )

            # Find a proper rigid transformation that produces the best combination
            # of atomic species in the Wyckoff positions.
            space_group = self.get_space_group_number()
            wyckoff_letters = self._get_spglib_wyckoff_letters_conventional()
            equivalent_atoms = self._get_spglib_equivalent_atoms_conventional()
            ideal_sys, ideal_wyckoff = self._find_wyckoff_ground_state(
                space_group,
                wyckoff_letters,
                spglib_conv_sys,
                is_flat=is_flat,
                nonperiodic_axis=nonperiodic_axis
            )

            # Center the system in the non-periodic direction, also taking
            # periodicity into account. Without the centering the structure may
            # end up being split at the cell boundary. The
            # get_center_of_mass()-function in MatID takes into account
            # periodicity and can produce the correct CM unlike the similar
            # function in ASE.
            ideal_sys.set_pbc(True)  # Needed temprorarily for centering to work
            pbc_cm = matid.geometry.get_center_of_mass(ideal_sys)
            cell_center = 0.5 * np.sum(ideal_sys.get_cell(), axis=0)
            translation = cell_center - pbc_cm
            conv_pbc = np.array([True, True, True])
            conv_pbc[nonperiodic_axis] = False
            translation[conv_pbc] = 0
            ideal_sys.translate(translation)
            ideal_sys.wrap()

            # Minimize the cell to only just fit the atoms in the non-periodic
            # direction
            min_conv_cell = matid.geometry.get_minimized_cell(
                ideal_sys,
                nonperiodic_axis,
                self.min_2d_thickness
            )

            # For the final system we set the correct pbc
            min_conv_cell.set_pbc(conv_pbc)

            self._conventional_system = min_conv_cell
            self._conventional_wyckoff_letters = ideal_wyckoff
            self._conventional_equivalent_atoms = equivalent_atoms
            return self._conventional_system
        else:
            raise ValueError("The provided system does not have 3 or 2 periodic directions.")

    def get_rotations(self):
        """Get the rotational parts of the Seitz matrices that are associated
        with this space group. Each rotational matrix is accompanied by a
        translation with the same index.

        Returns:
            np.ndarray: Rotation matrices.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["rotations"]

        return value

    def get_translations(self):
        """Get the translational parts of the Seitz matrices that are
        associated with this space group. Each translation is accompanied
        by a rotational matrix with the same index.

        Returns:
            np.ndarray: Translation vectors.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["translations"]

        return value

    def get_choice(self):
        """
        Returns:
            str: A string specifying the centring, origin and basis vector
            settings.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["choice"]

        return value

    def get_wyckoff_letters_original(self):
        """
        Returns:
            list of str: Wyckoff letters for the atoms in the original system.
        """
        spglib_wyckoffs = self._get_spglib_wyckoff_letters_original()
        if self._best_transform is None:
            self.get_conventional_system()
        permutations = self._best_transform["permutations"]
        new_wyckoffs = []
        for old_wyckoff in spglib_wyckoffs:
            new_wyckoff = permutations[old_wyckoff]
            new_wyckoffs.append(new_wyckoff)

        return np.array(new_wyckoffs)

    def get_equivalent_atoms_original(self):
        """
        The equivalent atoms are the same as what spglib already outputs, as
        changes in the wyckoff letters will not afect the equivalence:

        Returns:
            list of int: A list that maps each atom into a symmetry equivalent
                set.
        """
        spglib_equivalent_atoms = self._get_spglib_equivalent_atoms_original()
        return spglib_equivalent_atoms

    def get_wyckoff_letters_conventional(self):
        """Get the Wyckoff letters of the atoms in the conventional system.

        Returns:
            list of str: Wyckoff letters.
        """
        if self._conventional_wyckoff_letters is None:
            self.get_conventional_system()
        return self._conventional_wyckoff_letters

    def get_wyckoff_sets_conventional(self, return_parameters=True):
        """Get a list of Wyckoff sets for this system. Wyckoff sets combine
        information about the atoms and their positions at specific Wyckoff
        positions.

        Args:
            return_parameters (bool): Whether to return the value of possible
                free Wyckoff parameters. Set to false if they are not needed,
                as their determination can take some time.

        Returns:
            list of WyckoffSets: A list of :class:`.WyckoffSet` objects for the
            conventional system.
        """
        space_group = self.get_space_group_number()
        conv_sys = self.get_conventional_system()
        wyckoff_letters = self.get_wyckoff_letters_conventional()
        equivalent_atoms = self.get_equivalent_atoms_conventional()
        sets = self._get_wyckoff_sets(
            conv_sys,
            space_group,
            wyckoff_letters,
            equivalent_atoms,
            precision=0.5*self.symmetry_tol,
            return_parameters=return_parameters,
        )

        return sets

    def get_equivalent_atoms_conventional(self):
        """List of equivalent atoms in the idealized system.

        Returns:
            list of int: A list that maps each atom into a symmetry equivalent
                set.
        """
        if self._spglib_equivalent_atoms_conventional is None:
            self._get_spglib_wyckoff_letters_and_equivalent_atoms()
        return self._spglib_equivalent_atoms_conventional

    def get_wyckoff_letters_primitive(self):
        """Get the Wyckoff letters of the atoms in the primitive system.

        Returns:
            list of str: Wyckoff letters.
        """
        if self._primitive_wyckoff_letters is None:
            self.get_primitive_system()
        return self._primitive_wyckoff_letters

    def get_equivalent_atoms_primitive(self):
        """List of equivalent atoms in the primitive system.

        Returns:
            list of int: A list that maps each atom into a symmetry equivalent
                set.
        """
        if self._primitive_equivalent_atoms is None:
            self.get_primitive_system()
        return self._primitive_equivalent_atoms

    def get_symmetry_dataset(self):
        """Calculates the symmetry dataset with spglib for the given system.
        """
        if self._symmetry_dataset is not None:
            return self._symmetry_dataset

        description = self._system_to_spglib_description(self._analyzed_system)
        # Spglib has been observed to cause segmentation faults when fed with
        # invalid data, so run in separate process to catch those cases
        try:
            symmetry_dataset = segfault_protect(
                spglib.get_symmetry_dataset,
                description,
                self.symmetry_tol)
        except RuntimeError:
            raise CellNormalizationError(
                "Segfault in spglib when finding symmetry dataset. Please check "
                " the given cell, scaled positions and atomic numbers."
            )
        if symmetry_dataset is None:
            raise CellNormalizationError(
                'Spglib error when finding symmetry dataset.')

        self._symmetry_dataset = symmetry_dataset

        return symmetry_dataset

    def _get_spglib_conventional_system(self):
        """Returns an idealized description for this material as defined by
        spglib.

        Returns:
            ASE.Atoms: The idealized system as defined by spglib.
        """
        if self._spglib_conventional_system is not None:
            return self._spglib_conventional_system

        dataset = self.get_symmetry_dataset()
        cell = dataset["std_lattice"]
        pos = dataset["std_positions"]
        num = dataset["std_types"]
        spg_conv_sys = self._spglib_description_to_system((cell, pos, num))

        self._spglib_conventional_system = spg_conv_sys
        return spg_conv_sys

    def _get_spglib_wyckoff_letters_original(self):
        """
        Returns:
            list of str: Wyckoff letters for the atoms in the original system.
        """
        dataset = self.get_symmetry_dataset()
        value = np.array(dataset["wyckoffs"])

        return value

    def _get_spglib_equivalent_atoms_original(self):
        """
        Returns:
            list of int: A list that maps each atom into a symmetry equivalent
                set.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["equivalent_atoms"]

        return value

    def _get_spglib_wyckoff_letters_conventional(self):
        """
        Returns:
            np.array of str: Wyckoff letters for the atoms in the conventioal
            system as defined by spglib.
        """
        if self._spglib_wyckoff_letters_conventional is None:
            wyckoff_letters_primitive = self._get_spglib_wyckoff_letters_primitive()
            dataset = self.get_symmetry_dataset()
            mapping = dataset["std_mapping_to_primitive"]
            self._spglib_wyckoff_letters_conventional = wyckoff_letters_primitive[mapping]
        return self._spglib_wyckoff_letters_conventional

    def _get_spglib_equivalent_atoms_conventional(self):
        """
        Returns:
            np.array of int: List of numbers where the atoms are grouped to
            symmetrically equivalent groups by number.
        """
        if self._spglib_equivalent_atoms_conventional is None:
            equivalent_atoms_primitive = self._get_spglib_equivalent_atoms_primitive()
            dataset = self.get_symmetry_dataset()
            mapping = dataset["std_mapping_to_primitive"]
            self._spglib_equivalent_atoms_conventional = equivalent_atoms_primitive[mapping]

        return self._spglib_equivalent_atoms_conventional

    def _get_spglib_origin_shift(self):
        """The origin shift s that is needed to transform points in the
        original system to the conventional system. The relation between the
        original coordinates and the conventional coordinates is defined by:

        x' = P*x + s

        where x' is a coordinate in the conventional system, P is the
        transformation matrix, x is a coordinate in the original system and s
        is the origin shift.

        Returns:
            3*1 np.ndarray: The shift of the origin as a vector.
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["origin_shift"]

        return value

    def get_symmetry_operations(self):
        """The symmetry operations of the original structure as rotations and
        translations.

        Returns:
            Dictionary containing an entry for rotations containing a np.array
            with 3x3 matrices for each symmetry operation and an entry
            "translations" containing np.array of translations for each
            symmetry operation.
            3*1 np.ndarray: The shift of the origin as a vector.
        """
        dataset = self.get_symmetry_dataset()
        operations = {
            "rotations": dataset["rotations"],
            "translations": dataset["translations"]
        }

        return operations

    def _get_spglib_transformation_matrix(self):
        """The transformation matrix P that transforms points in the original
        system to the conventional system. The relation between the original
        coordinates and the conventional coordinates is defined by:

        x' = P*x + s

        where x' is a coordinate in the conventional system, P is the
        transformation matrix, x is a coordinate in the original system and s
        is the origin shift.

        Returns:
            3x3 np.ndarray:
        """
        dataset = self.get_symmetry_dataset()
        value = dataset["transformation_matrix"]

        return value

    def _get_spglib_primitive_system(self):
        """Returns a primitive description as defined by spglib for this
        system.

        Returns:
            ASE.Atoms: The primitive system.
        """
        if self._spglib_primitive_system is not None:
            return self._spglib_primitive_system

        spglib_conv_sys = self._get_spglib_conventional_system()
        spglib_conv_wyckoff = self._get_spglib_wyckoff_letters_conventional()
        spglib_conv_equivalent = self._get_spglib_equivalent_atoms_conventional()
        space_group_short = self.get_space_group_international_short()

        spglib_prim_sys, spglib_prim_wyckoff, spglib_prim_equivalent = self._get_primitive_system(
            spglib_conv_sys, spglib_conv_wyckoff, spglib_conv_equivalent, space_group_short)

        self._spglib_primitive_system = spglib_prim_sys
        self._spglib_wyckoff_letters_primitive = spglib_prim_wyckoff
        self._spglib_equivalent_atoms_primitive = spglib_prim_equivalent

        return self._spglib_primitive_system

    def _get_spglib_wyckoff_letters_primitive(self):
        """Get the Wyckoff letters of the atoms in the primitive system as
        defined by spglib.

        Returns:
            list of str: Wyckoff letters.
        """
        if self._spglib_wyckoff_letters_primitive is None:
            wyckoff_letters_original = self._get_spglib_wyckoff_letters_original()
            mapping = self._get_spglib_primitive_to_original_mapping()
            self._spglib_wyckoff_letters_primitive = wyckoff_letters_original[mapping]
        return self._spglib_wyckoff_letters_primitive

    def _get_spglib_equivalent_atoms_primitive(self):
        """List of equivalent atoms in the primitive system as defined by
        spglib.

        Returns:
            list of int: A list that maps each atom into a symmetry equivalent
                set.
        """
        if self._spglib_equivalent_atoms_primitive is None:
            equivalent_atoms_original = self._get_spglib_equivalent_atoms_original()
            mapping = self._get_spglib_primitive_to_original_mapping()
            self._spglib_equivalent_atoms_primitive = equivalent_atoms_original[mapping]
        return self._spglib_equivalent_atoms_primitive

    def _get_spglib_primitive_to_original_mapping(self):
        """Returns a mapping from that links an atom in the primitive cell to
        one of the duplicates in the original system.

        Returns:
            np.ndarray: A list of integer indices, one for each atom in the
            primitive system as returned by spglib. The indices refer to an
            atom in the original simulation system.
        """
        if self._spglib_primitive_to_original_mapping is None:
            dataset = self.get_symmetry_dataset()
            mapping = dataset["mapping_to_primitive"]
            _, indices = np.unique(mapping, return_index=True)
            self._spglib_primitive_to_original_mapping = indices

        return self._spglib_primitive_to_original_mapping

    def _get_primitive_system(
            self,
            conv_system,
            conv_wyckoff,
            conv_equivalent,
            space_group_international_short):
        """Returns an primitive description for an idealized system in the
        conventional cell. This description uses a primitive lattice
        where positions of the atoms, and the cell basis vectors are idealized
        to follow the symmetries that were found with the given precision. This
        means that e.g. the volume, angles between basis vectors and basis
        vector lengths may have deviations from the the original system.

        The transformation matrices from conventional system to primitive are
        as given at: https://atztogo.github.io/spglib/definition.html#id8

        Args:
            conv_system (ase.Atoms): The conventional system from which the
                primitive system is created.
            conv_wyckoff (np.array of str): Wyckoff letters of the given
                conventional system
            conv_equivalent (np.array of int): Equivalent atoms of the given
                conventional system
            space_group_international_short (str): The space group symbol in
                international short form

        Returns:
            tuple containing ase.Atoms, wyckoff_letters and equivalent atoms
        """
        centring = space_group_international_short[0]

        # For the primitive centering the conventional lattice is the primitive
        # as well
        if centring == "P":
            return conv_system, conv_wyckoff, conv_equivalent

        primitive_transformations = {
            "A": np.array([
                [1, 0, 0],
                [0, 1/2, -1/2],
                [0, 1/2, 1/2],
            ]),
            "C": np.array([
                [1/2, 1/2, 0],
                [-1/2, 1/2, 0],
                [0, 0, 1],
            ]),
            "R": np.array([
                [2/3, -1/3, -1/3],
                [1/3, 1/3, -2/3],
                [1/3, 1/3, 1/3],
            ]),
            "I": np.array([
                [-1/2, 1/2, 1/2],
                [1/2, -1/2, 1/2],
                [1/2, 1/2, -1/2],
            ]),
            "F": np.array([
                [0, 1/2, 1/2],
                [1/2, 0, 1/2],
                [1/2, 1/2, 0],
            ]),
        }

        # Transform conventional cell to the primitive cell
        transform = primitive_transformations[centring]
        conv_cell = conv_system.get_cell()
        prim_cell = np.dot(transform.T, conv_cell)

        # Transform all position to the basis of the primitive cell
        conv_pos = conv_system.get_positions()
        prim_cell_inv = np.linalg.inv(prim_cell)
        prim_pos = np.dot(conv_pos, prim_cell_inv)

        # Keep one occurrence for each atom that should be within the cell and
        # wrap it's position to tbe inside the primitive cell.
        conv_num = conv_system.get_atomic_numbers()
        conv_to_prim_map = self._symmetry_dataset["std_mapping_to_primitive"]
        _, inside_mask = np.unique(conv_to_prim_map, return_index=True)
        prim_pos = prim_pos[inside_mask]
        prim_num = conv_num[inside_mask]

        # Store the wyckoff letters and equivalent atoms
        prim_wyckoff = conv_wyckoff[inside_mask]
        prim_equivalent = conv_equivalent[inside_mask]

        prim_sys = Atoms(
            scaled_positions=prim_pos,
            symbols=prim_num,
            cell=prim_cell,
        )
        prim_sys.wrap(pbc=True)

        return prim_sys, prim_wyckoff, prim_equivalent

    def _system_to_spglib_description(self, system):
        """Transforms the given ASE.Atoms object into a tuple used by spglib.
        """
        angstrom_cell = self._analyzed_system.get_cell()
        relative_pos = self._analyzed_system.get_scaled_positions()
        atomic_numbers = self._analyzed_system.get_atomic_numbers()
        description = (angstrom_cell, relative_pos, atomic_numbers)

        return description

    def _spglib_description_to_system(self, desc):
        """Transforms a tuple used by spglib into ASE.Atoms
        """
        system = Atoms(
            numbers=desc[2],
            cell=desc[0],
            scaled_positions=desc[1],
        )

        return system

    def _wrap_positions(self, positions, precision=1E-5, copy=True):
        """Wrap positions so that each element in the array is within the
        half-closed interval [0, 1)

        By wrapping values near 1 to 0 we will have a consistent way of
        presenting systems.

        Args:
            positions (np.ndarray): Atomic positions that are given in the unit cell basis.
            precision (float): The precision for wrapping coordinates that are close to

                zero or unity.
            copy (bool): Whether a the returned value is a copy or the values are
                modified in-place.

        Returns:
            np.ndarray: The new wrapped positions.
        """
        if copy:
            wrapped_positions = np.copy(positions)
        else:
            wrapped_positions = positions

        wrapped_positions %= 1
        abs_zero = np.absolute(wrapped_positions)
        abs_unity = np.absolute(abs_zero-1)

        near_zero = np.where(abs_zero < precision)
        near_unity = np.where(abs_unity < precision)

        wrapped_positions[near_unity] = 0
        wrapped_positions[near_zero] = 0

        return wrapped_positions

    def _search_periodic_positions(
            self,
            target_pos,
            positions,
            cell,
            accuracy):
        """Searches a list of positions for a match for the target position taking
        into account the periodicity of the system.

        Args:
            target_pos (1x3 np.array): The relative position to search.
            positions (Nx3 np.array): The relative position where to search.
            cell (3x3 np.array): The cell used to find a threshold accuracy in
                cartesian coordinates.
            accuracy (float): The minimum cartesian distance (angstroms) that is
                required for the atoms to be considered identical.

        Returns:
            If a match is found, returns the index of the match in 'positions'. If
            no match is found, returns None.
        """
        if len(positions.shape) == 1:
            positions = positions[np.newaxis, :]

        # Calculate the distances without taking into account periodicity.
        # Here we calculate all the distances although in reality we could loop
        # over the distances and calculate only until the correct one is found.
        # But it turns out it is faster to calculate the distances i one
        # vectorized operation with numpy than a python loop.
        displacements = positions-target_pos

        # Take periodicity into account by wrapping coordinate elements that
        # are bigger than 0.5 or smaller than -0.5
        indices = np.where(displacements > 0.5)
        displacements[indices] = displacements[indices] - 1
        indices = np.where(displacements < -0.5)
        displacements[indices] = displacements[indices] + 1

        # Convert displacements to cartesian coordinates
        displacements = np.dot(displacements, cell.T)

        # Try to find a match for the target by finding a distance that is less
        # than the accuracy
        distances = np.linalg.norm(displacements, axis=1)

        min_index = np.argmin(distances)
        min_distance = distances[min_index]
        if min_distance <= accuracy:
            return min_index
        else:
            return None

    def _find_wyckoff_ground_state(
            self,
            space_group,
            old_wyckoff_letters,
            system,
            is_flat=False,
            nonperiodic_axis=None):
        """
        When given a system that has been normalized by spglib, this function
        will find a atomic positions within that cell that are most unique
        (totally unique up to isotropic scaling if no free Wyckoff parameters
        present).

        The function is based on finding a "normalizer" (found for each space
        group e.g. at the Bilbao Crystallographic Server)
        http://www.cryst.ehu.es/), which is essentially a transform that
        changes Wyckoff positions of atoms within a cell without breaking the
        symmetry. Each of these normalizers, that corresponds to a proper rigid
        transformation in the cartesian basis, is applied to give a different
        structural representation. The algorithm then goes through each tuple
        of Wyckoff letter and atomic number (W , Z) in a preset order: the
        first loop goes through the Wyckoff letters in alphabetical order, and
        the second loop goes through the atomic numbers from lowest to highest.
        Whenever some of the possible representations has a structural
        component corresponding to the current tuple (W , Z), the number of
        atoms with this tuple N is calculated. The representation is stored to
        a map structure that links each N to a list of representations and the
        highest N is tracked. After all the representations are covered, the
        candidate list of representations is replaced with the list
        corresponding to the highest N. The algorithm stops whenever the
        candidate set contains only one representation, which will be the
        standard one.

        Args:
            space_group(int): The space group of the system.
            old_wyckoff_letters(list of strings): Wyckoff letters as detected
                by spglib for the atoms in the given system.
            system(ase.Atoms): The standardized system as given by spglib.
            is_flat(bool): Whether the structure is flat (near zero thickness)
                in one non-periodic direction. Applies only for 2D systems.
            nonperiodic_axis(int): The index of a nonperiodic axis in the cell
                basis. Applies only for 2D systems.

        Returns:
            (ase.Atoms, list of strings): Returns a tuple containing the found
            conventional system and the Wyckoff letters for it.
        """
        # Gather the allowed transformations. For completely flat structures (all
        # atoms in 2D plane), also the rigid transformation that are improper
        # (determinant -1), will become proper, as we can always invert the
        # non-periodic axis to change the sign of the determinant. Physically this
        # corresponds to rotating the system rigidly through the third nonperiodic
        # dimension.
        transform_list = []
        identity = {
            "transformation": np.identity(4),
            "permutations": {x: x for x in old_wyckoff_letters},
            "identity": True,
        }
        transform_list.append(identity)

        proper_rigid_trans = PROPER_RIGID_TRANSFORMATIONS.get(space_group)
        if proper_rigid_trans is not None:
            transform_list.extend(proper_rigid_trans)
        improper_rigid_trans = IMPROPER_RIGID_TRANSFORMATIONS.get(space_group)
        if is_flat:
            improper_rigid_trans = IMPROPER_RIGID_TRANSFORMATIONS.get(space_group)
            if improper_rigid_trans is not None:
                transform_list.extend(improper_rigid_trans)

        # Test which transformations are proper rigid transformation for the
        # current cell. TODO: Could the proper rigid transformation be checked
        # beforehand for each normalizer by looking at the crystal lattice
        # characteristics like orthogonality, basis sizes, etc.?
        # TODO: There is an optimization that could slightly speed up the
        # calculation of whether the matrix is proper rigid: If one would
        # precalculate the inverse and transpose of the transformation matrices
        # in the scaled basis, then one can use the rules det(ABC) =
        # det(A)*det(B)*det(C), (ABC)^-1 = C^-1*B^-1*A^-1 and (ABC)^T =
        # C^T*B^T*A^T to speed up the calculation.
        cart_basis = np.eye(3)
        cell_basis = system.get_cell()

        # If the structure is flat, we ignore the non-periodic basis
        if nonperiodic_axis is not None:
            dim_mask = np.array((True, True, True))
            dim_mask[nonperiodic_axis] = False
            cart_basis = cart_basis[dim_mask, :]
            cart_basis = cart_basis[:, dim_mask]
            cell_basis = cell_basis[dim_mask, :]
            cell_basis = cell_basis[:, dim_mask]

        # These are the change of basis matrices for going from cartesian basis
        # to scaled basis and vice versa
        cart_to_cell = np.dot(cart_basis, np.linalg.inv(cell_basis))
        cell_to_cart = np.dot(cell_basis, np.linalg.inv(cart_basis))
        proper_transforms = []

        for trans_info in transform_list:
            trans = trans_info["transformation"]
            nonaugmented_trans = trans[0:3, 0:3]

            # Remove the non-periodic dimension from the transform as well
            if nonperiodic_axis is not None:
                nonaugmented_trans = nonaugmented_trans[dim_mask, :]
                nonaugmented_trans = nonaugmented_trans[:, dim_mask]

            # Here we transform the basis of the transformation matrix from
            # scaled coordinates to cartesian coordinates. This is necessary
            # because the scaled coordinates have a distorted metric and only
            # the cartesian version can reveal if the transformation is proper
            # rigid.
            cart_trans = np.dot(cart_to_cell, np.dot(nonaugmented_trans, cell_to_cart))

            # Check if the transformation is proper
            determinant = np.linalg.det(cart_trans)
            is_proper = abs(determinant - 1) < 1e-8

            # Check if transformation is orthogonal
            if is_proper:
                test_inv = np.linalg.inv(cart_trans)
                test_trans = cart_trans.T
                is_orthogonal = np.allclose(test_inv, test_trans, rtol=0, atol=1e-8)

                if is_orthogonal:
                    proper_transforms.append(trans_info)

        transform_list = proper_transforms

        # If no transformation found for this space group, return the same
        # system
        if len(transform_list) == 1:
            self._best_transform = identity
            return system, old_wyckoff_letters

        # Form all available representations
        representations = []
        atomic_numbers = system.get_atomic_numbers()
        for transform in transform_list:
            perm = transform["permutations"]
            representation = {
                "transformation": transform["transformation"],
                "permutations": perm,
            }
            wyckoff_positions = {}
            wyckoff_letters = []
            i_perm = 0
            for i_atom, old_w in enumerate(old_wyckoff_letters):
                new_w = perm.get(old_w)
                wyckoff_letters.append(new_w)
                if new_w is not None:
                    z = atomic_numbers[i_atom]
                    old_n_atoms = wyckoff_positions.get((new_w, z))
                    if old_n_atoms is None:
                        wyckoff_positions[(new_w, z)] = 1
                    else:
                        wyckoff_positions[(new_w, z)] += 1
                    i_perm += 1
            representation["wyckoff_positions"] = wyckoff_positions
            representations.append(representation)

        # Gather all available Wyckoff letters in all representations
        wyckoff_letters = set()
        for transform in transform_list:
            i_perm = transform["permutations"]
            for orig, new in i_perm.items():
                wyckoff_letters.add(new)
        wyckoff_letters = sorted(wyckoff_letters)

        # Gather all available atomic numbers
        atomic_numbers = set(system.get_atomic_numbers())
        atomic_numbers = sorted(atomic_numbers)

        # Decide the best representation
        best_representation = None
        found = False
        for w in wyckoff_letters:
            if found:
                break
            for z in atomic_numbers:
                n_atoms_map = defaultdict(list)
                n_atoms_max = 0
                for r in representations:
                    i_n = r["wyckoff_positions"].get((w, z))
                    if i_n is not None:
                        n_atoms_map[i_n].append(r)
                        if i_n > n_atoms_max:
                            n_atoms_max = i_n
                if n_atoms_max != 0:
                    representations = n_atoms_map[n_atoms_max]
                if len(representations) == 1:
                    best_representation = representations[0]
                    found = True

        # If no best transformation was found, then multiple transformation are
        # equal. Ensure this and then choose the first one.
        error = SystaxError("Could not successfully decide best Wyckoff positions.")
        if len(representations) > 1:
            new_wyckoffs = representations[0]["wyckoff_positions"]
            n_items = len(new_wyckoffs)
            for representation in representations[1:]:
                i_wyckoffs = representation["wyckoff_positions"]
                if len(i_wyckoffs) != n_items:
                    raise error
                for key in new_wyckoffs.keys():
                    if i_wyckoffs[key] != new_wyckoffs[key]:
                        raise error
        best_representation = representations[0]

        # Apply the best transform
        new_system = system.copy()
        if best_representation.get("identity"):
            self._best_transform = identity
            return new_system, old_wyckoff_letters
        else:
            self._best_transform = best_representation
            best_transformation_matrix = best_representation["transformation"]
            best_permutations = best_representation["permutations"]
            new_wyckoff_letters = []
            for i_atom, old_w in enumerate(old_wyckoff_letters):
                new_w = best_permutations.get(old_w)
                new_wyckoff_letters.append(new_w)
            new_wyckoff_letters = np.array(new_wyckoff_letters)

            # Create the homogeneus coordinates
            n_pos = len(system)
            old_pos = np.empty((n_pos, 4))
            old_pos[:, 3] = 1
            old_pos[:, 0:3] = system.get_scaled_positions()

            # Apply transformation with the augmented 3x4 matrix that is used
            # for homogeneous coordinates
            transformed_positions = np.dot(old_pos, best_transformation_matrix.T)

            # Get rid of the extra dimension of the homogeneous coordinates
            transformed_positions = transformed_positions[:, 0:3]

            # Wrap the positions to the half-closed interval [0, 1)
            wrapped_pos = matid.geometry.get_wrapped_positions(transformed_positions)
            new_system.set_scaled_positions(wrapped_pos)

            return new_system, new_wyckoff_letters

    def _get_wyckoff_sets(
            self,
            system,
            space_group,
            wyckoff_letters,
            equivalent_atoms,
            precision,
            return_parameters
            ):
        """Used to get detailed information about about the sets of equivalent
        atoms. The detected Wyckoff set variables (x, y, z) are reported
        consistenly by selecting the ariable sets that has lowest x value, then
        lowest y and finally lowest z.

        If return_parameters is set to True, the possible variables for the
        Wyckoff sets are returned. Because spglib does not currently print out
        detailed information about free parameters for Wyckoff sets, we use
        information from the Bilbao Crystallographic Database to get the values
        of the free variables for each group of symmetry related atoms. The
        positions R can be calculated as (using row vectors):

        .. math::
            \mathbf{R} = \mathbf{W}\mathbf{M} + \mathbf{C}

        where :math:`\mathbf{W}` is a row vector containing the Wyckoff variables,
        :math:`\mathbf{M}` is a matrix defining the multipliers for each Wyckoff
        variable and :math:`\mathbf{C}` is a vector containing the constant
        offsets.  From this expression the Wyckoff variables :math:`\mathbf{W}`
        may be solved as:

        .. math::
            \mathbf{W} = \mathbf{R-C}\mathbf{M}^{+}

        where :math:`\mathbf{M}^{+}` is the pseudo-inverse of matrix
        :math:`\mathbf{M}` (the set of linear equations is not always linearly
        independent, so regular inverse does not work).

        Args:
            system (System): The atomic system in which the atomic positions
                are inspected.
            space_group (int): The space group number
            wyckoff_letters (list): List of Wyckoff letters for each atom in
                the system
            equivalent_atoms (list): List of integers that distinguish atoms
                that are related by symmetry.
            precision (float): The precision for matching atoms to Wyckoff
                positions.
            return_parameters (bool): Whether to return the value of possible
                free Wyckoff parameters. Set to false if the are needed, as
                their determination can take time.

        Returns:
            list of WyckoffSets: A list of :class:`.WyckoffSet` objects for this
            system.
        """
        cell = system.get_cell()
        elements = system.get_chemical_symbols()
        numbers = system.get_atomic_numbers()
        positions = system.get_scaled_positions()
        wyckoff_infos = WYCKOFF_SETS[space_group]

        # Form the Wyckoff sets
        sets = {}
        unique_sets, unique_indices = np.unique(
            equivalent_atoms,
            return_index=True
        )
        for i_set, index in enumerate(unique_indices):
            set_index = unique_sets[i_set]

            set_data = WyckoffSet(
                wyckoff_letter=str(wyckoff_letters[index]),
                element=str(elements[index]),
                atomic_number=int(numbers[index]),
                space_group=space_group,
                representative=wyckoff_infos[wyckoff_letters[index]]["expressions"][0],
            )
            set_data.indices = []
            sets[set_index] = set_data

        # Add the indices of the atoms that belong to these sets, and add the
        # multiplicity of the set
        for i_atom, set_number in enumerate(equivalent_atoms):
            sets[set_number].indices.append(i_atom)
        for wset in sets.values():
            wset.multiplicity = len(wset.indices)

        # For each set, solve the free variables if any present.
        if return_parameters:
            translations = wyckoff_infos["translations"]
            for i_set, wset in sets.items():
                indices = wset.indices[:]
                wyckoff_letter = wset.wyckoff_letter
                wyckoff_info = wyckoff_infos[wyckoff_letter]
                variables_present = wyckoff_info["variables"]
                all_pos = positions[indices]

                # Get the precalculated matrices and vectors that are needed
                # for solving the system of linear equations. 
                Ms = wyckoff_info["matrices"]
                M_pinvs = wyckoff_info["pinv_matrices"]
                Cs = wyckoff_info["constants"]

                # Resolve the needed variables
                if variables_present:
                    n_expr = Cs.shape[0]
                    n_trans = len(translations)
                    W_final = None

                    # Sort the positions to find a consistent set of variables
                    # each time. The positions are ordered by x, then y, and
                    # finally z. Another option would have been to find all
                    # possible  Wyckoff variable alternatives and instead sort
                    # them to select one. That however is much more
                    # time-consuming for larger, highly symmemric structures
                    # with many variable alternatives.
                    sorted_indices = np.lexsort(np.flip(all_pos.T, axis=0))
                    sorted_pos = all_pos[sorted_indices]

                    # Calculate the variables (x, y, z) based on the first
                    # Wyckoff position. The variables are calculated for each
                    # atom, and they are stored in the list of valid values if
                    # the other generated positions can be matched with the
                    # rest of the atoms.
                    variable_candidates = []
                    for atom_index in indices:

                        # Calculate the Wyckoff variables based on the atom at
                        # current index and the first (and often the simplest)
                        # Wyckoff position expression.
                        R = positions[atom_index]
                        W = np.dot(R - Cs[0], M_pinvs[0])

                        # Calculate the positions of all other atoms with the
                        # currently tested Wyckoff variables
                        test_positions = np.zeros(((n_trans+1)*n_expr, 3))
                        first_test_pos = np.dot(W, Ms) + Cs
                        test_positions[0:n_expr, :] = first_test_pos
                        i_trans = 1
                        for trans in translations:
                            test_positions[i_trans*n_expr:(i_trans+1)*n_expr, :] = first_test_pos + trans
                            i_trans += 1
                        test_positions = matid.geometry.get_wrapped_positions(test_positions)

                        # Test if each test positions can be matched to an atom
                        # in the actual structure. If yes, save this set of
                        # variables and break the loop.
                        found = True
                        for test_pos in test_positions:
                            if self._search_periodic_positions(
                                    test_pos,
                                    sorted_pos,
                                    cell,
                                    precision) is None:
                                found = False
                                break
                        if found:
                            W_final = W
                            break

                    if W_final is None:
                        raise ValueError(
                            "Could not resolve the free Wyckoff parameters for "
                            "Wyckoff letter '{}' in space group {}. Problem in "
                            "determining variables for element '{}' at indices "
                            "'{}'."
                            .format(wset.wyckoff_letter, wset.space_group, wset.element, wset.indices)
                        )
                    else:
                        # Wrap Wyckoff variables to be between [0, 1] and save
                        # to the WyckoffSet
                        W_final = matid.geometry.get_wrapped_positions(W_final)
                        for j, var in enumerate(["x", "y", "z"]):
                            if var in variables_present:
                                setattr(wset, var, W_final[j])

        # Sort the list so that sets with Wyckoff letter earlier in the
        # alphabet are first, and sets with the same Wyckoff letter are
        # sorted by atomic number. Groups with the same Wyckoff letter and
        # atomic number are still randomly sorted.
        unsorted_list = list(sets.values())
        sorted_list = sorted(unsorted_list, key=attrgetter('wyckoff_letter', 'atomic_number'))

        return sorted_list



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import super
__metaclass__ = type


class Classification():
    def __init__(self, atoms):
        self.atoms = atoms


#===============================================================================
# 0D Structures
class Class0D(Classification):
    """Structures that have a structure that is isolated in all directions by a
    vacuum gap.
    """


class Atom(Class0D):
    """
    """


#===============================================================================
# 1D Structures
class Class1D(Classification):
    """All structures that are roughly 1-dimensional, meaning that one
    dimension is much larger than the two others.
    """


#===============================================================================
# 2D Structures
class Class2D(Classification):
    """Base class for all structures that are roughly 2-dimensional, meaning that two of the
    dimensions are much larger than the remaining one.
    """


class Class2DWithCell(Class2D):
    """Two dimensional structures from which a periodic unit cell has been
    identified.
    """
    def __init__(
            self,
            atoms,
            region,
            ):
        super().__init__(atoms)
        self.region = region

    @property
    def basis_indices(self):
        return self.region.get_basis_indices()

    @property
    def outliers(self):
        return self.region.get_outliers()

    @property
    def interstitials(self):
        return self.region.get_interstitials()

    @property
    def adsorbates(self):
        return self.region.get_adsorbates()

    @property
    def substitutions(self):
        return self.region.get_substitutions()

    @property
    def vacancies(self):
        return self.region.get_vacancies()

    @property
    def unknowns(self):
        return self.region.get_unknowns()

    @property
    def prototype_cell(self):
        return self.region.cell


class Surface(Class2DWithCell):
    """
    """


class Material2D(Class2DWithCell):
    """
    """


#===============================================================================
# 3D Structures
class Class3D(Classification):
    """All structures that periodically extend infinitely without vacuum gaps.
    """


#===============================================================================
# Unknown structures
class Unknown(Classification):
    """
    """



"""Utility functions to protect main process against segfaults"""
from __future__ import absolute_import, division, print_function, unicode_literals
import signal


def sig_handler(signum, frame):
    raise RuntimeError("SIGSEGV encountered.")


def segfault_protect(function, *args, **kwargs):
    """Used to run a function in a separate process to catch RuntimeErrors such
    as segfaults.
    """
    signal.signal(signal.SIGSEGV, sig_handler)
    result = function(*args)
    return result



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import (bytes, str, open, super, range,
                      zip, round, input, int, pow, object)

import numpy as np

import chronic

# from ase.visualize import view
from ase.data import covalent_radii

from matid.classifications import \
    Surface, \
    Atom, \
    Material2D, \
    Unknown, \
    Class0D, \
    Class1D, \
    Class2D, \
    Class3D
    # Crystal
import matid.geometry
from matid.data import constants
from matid.classification.periodicfinder import PeriodicFinder
# from matid.symmetry.symmetryanalyzer import SymmetryAnalyzer

__metaclass__ = type


class Classifier():
    """A class that is used to analyze the contents of an atomistic system and
    separate the consituent atoms into different components along with other
    meaningful information.
    """
    def __init__(
            self,
            seed_position="cm",
            max_cell_size=constants.MAX_CELL_SIZE,
            pos_tol=None,
            pos_tol_mode="relative",
            pos_tol_scaling=constants.POS_TOL_SCALING,
            angle_tol=constants.ANGLE_TOL,
            cluster_threshold=constants.CLUSTER_THRESHOLD,
            crystallinity_threshold=constants.CRYSTALLINITY_THRESHOLD,
            delaunay_threshold=constants.DELAUNAY_THRESHOLD,
            bond_threshold=constants.BOND_THRESHOLD,
            delaunay_threshold_mode="relative",
            chem_similarity_threshold=constants.CHEM_SIMILARITY_THRESHOLD,
            cell_size_tol=constants.CELL_SIZE_TOL,
            max_n_atoms=constants.MAX_N_ATOMS,
            max_2d_cell_height=constants.MAX_2D_CELL_HEIGHT,
            max_2d_single_cell_size=constants.MAX_SINGLE_CELL_SIZE,
            symmetry_tol=constants.SYMMETRY_TOL,
            min_coverage=constants.MIN_COVERAGE
            ):
        """
        Args:
            seed_position(str or np.ndarray): The seed position. Either provide
                a 3D vector from which the closest atom will be used as a seed, or
                then provide a valid option as a string. Valid options are:
                    - 'cm': One seed nearest to center of mass
            max_cell_size(float): The maximum size of cell basis vectors.
            pos_tol(float): The position tolerance for finding translationally
                repeated units. The units depend on 'pos_tol_mode'.
            pos_tol_mode(str): The mode for calculating the position tolerance.
                One of the following:
                    - "relative": Tolerance relative to the average nearest
                      neighbour distance.
                    - "absolute": Absolute tolerance in angstroms.
            pos_tol_scaling(float): The distance dependent scaling factor for
                the positions tolerance.
            angle_tol(float): The angle below which vectors in the cell basis are
                considered to be parallel.
            cluster_threshold(float): A parameter that controls which atoms are
                considered to be energetically connected when clustering is
                perfomed. Given in angstroms.
            crystallinity_threshold(float): The threshold of number of symmetry
                operations per atoms in primitive cell that is required for
                crystals.
            max_n_atoms(int): The maximum number of atoms in the system. If the
                system has more atoms than this, a ValueError is raised. If
                undefined, there is no maximum.
            bond_threshold(float): The clustering threshold when determining
                the connectivity of atoms in a surface or 2D-material.
            delaunay_threshold(str): The maximum length of an edge in the
                Delaunay triangulation.
            delaunay_threshold_mode(str): The mode for calculating the maximum
                length of an edge in the Delaunay triangulation.
                One of the following:
                    - "relative": Tolerance relative to the average nearest
                      neighbour distance.
                    - "absolute": Absolute tolerance in angstroms.
            pos_tol_factor(float): The factor for multiplying the position
                tolerance when searching neighbouring cell seed atoms.
            cell_size_tol(float): The tolerance for cell sizes to be considered
                equal. Given relative to the smallest cell size.
            max_2d_cell_height(float): The maximum allowed thickness in for a 2D
                material. Given in angstroms.
            max_2d_single_cell_size(float): The maximum allowed cell size for
                2D materials with only one unit cell found in the simulation
                cell. Given in angstroms.
            symmetry_tol(float): The tolerance for finding symmetry positions
                when determining the conventional cell for bulk structures. Given
                in angstroms.
            min_coverage(float): The minimum fraction that a found periodic
                region has to cover in the structure for the entire structure to be
                classified based on that found region (as surface or 2D
                material).
        """
        if pos_tol_mode == "relative" and pos_tol is None:
            pos_tol = constants.REL_POS_TOL
        if isinstance(max_cell_size, (int, float)):
            max_cell_size = [max_cell_size]
        self.max_cell_size = max_cell_size
        if isinstance(pos_tol, (int, float)):
            pos_tol = [pos_tol]
        self.pos_tol = pos_tol
        self.pos_tol_scaling = pos_tol_scaling
        self.abs_pos_tol = None
        self.pos_tol_mode = pos_tol_mode
        self.angle_tol = angle_tol
        self.crystallinity_threshold = crystallinity_threshold
        self.cluster_threshold = cluster_threshold
        self.delaunay_threshold = delaunay_threshold
        self.abs_delaunay_threshold = None
        self.delaunay_threshold_mode = delaunay_threshold_mode
        self.bond_threshold = bond_threshold
        self.chem_similarity_threshold = chem_similarity_threshold
        self.pos_tol_scaling = pos_tol_scaling
        self.cell_size_tol = cell_size_tol
        self.max_n_atoms = max_n_atoms
        self.max_2d_cell_height = max_2d_cell_height
        self.max_2d_single_cell_size = max_2d_single_cell_size
        self.symmetry_tol = symmetry_tol
        self.min_coverage = min_coverage

        # Check seed position
        if type(seed_position) == str:
            if seed_position == "cm":
                pass
            else:
                raise ValueError(
                    "Unknown seed_position: '{}'. Please provide a 3D vector "
                    "or a valid option as a string.".format(seed_position)
                )
        self.seed_position = seed_position

        # Check pos tolerance mode
        allowed_modes = set(["relative", "absolute"])
        if pos_tol_mode not in allowed_modes:
            raise ValueError("Unknown value '{}' for 'pos_tol_mode'.".format(pos_tol_mode))

        # Check delaunay tolerance mode
        allowed_modes = set(["relative", "absolute"])
        if delaunay_threshold_mode not in allowed_modes:
            raise ValueError("Unknown value '{}' for 'delaunay_threshold_mode'.".format(delaunay_threshold_mode))

    def classify(self, input_system):
        """A function that analyzes the system and breaks it into different
        components.

        Args:
            system(ASE.Atoms or System): Atomic system to classify.

        Returns:
            Classification: One of the subclasses of the Classification base
            class that represents a classification.

        Raises:
            ValueError: If the system has more atoms than self.max_n_atoms
        """
        # We wrap the positions to to be inside the cell.
        system = input_system.copy()
        system.wrap()
        self.system = system
        classification = None

        n_atoms = len(system)
        if n_atoms > self.max_n_atoms:
            raise ValueError(
                "The system contains more atoms ({}) than the current allowed "
                "limit of {}. If you wish you can increase this limit with the "
                "max_n_atoms attribute.".format(n_atoms, self.max_n_atoms)
            )

        # Calculate the displacement tensor for the original system. It will be
        # reused in multiple sections.
        pos = system.get_positions()
        cell = system.get_cell()
        pbc = system.get_pbc()

        with chronic.Timer("displacement_tensor"):
            disp_tensor = matid.geometry.get_displacement_tensor(pos, pos)
            if pbc.any():
                disp_tensor_pbc, disp_factors = matid.geometry.get_displacement_tensor(
                    pos,
                    pos,
                    cell,
                    pbc,
                    mic=True,
                    return_factors=True
                )
            else:
                disp_tensor_pbc = disp_tensor
                disp_factors = np.zeros(disp_tensor.shape)
            dist_matrix_pbc = np.linalg.norm(disp_tensor_pbc, axis=2)

        # Calculate the distance matrix where the periodicity and the covalent
        # radii have been taken into account
        dist_matrix_radii_pbc = np.array(dist_matrix_pbc)
        num = system.get_atomic_numbers()
        radii = covalent_radii[num]
        radii_matrix = radii[:, None] + radii[None, :]
        dist_matrix_radii_pbc -= radii_matrix

        # If pos_tol_mode or delaunay_threshold_mode is relative, get the
        # average distance to closest neighbours
        if self.pos_tol_mode == "relative" or self.delaunay_threshold_mode == "relative":
            min_basis = np.linalg.norm(cell, axis=1).min()
            dist_matrix_mod = np.array(dist_matrix_pbc)
            np.fill_diagonal(dist_matrix_mod, min_basis)
            global_min_dist = dist_matrix_mod.min()
            min_dist = np.min(dist_matrix_mod, axis=1)
            mean_min_dist = min_dist.mean()

            if self.pos_tol_mode == "relative":
                self.abs_pos_tol = np.array(self.pos_tol)*global_min_dist
            elif self.pos_tol_mode == "absolute":
                self.abs_pos_tol = self.pos_tol

            if self.delaunay_threshold_mode == "relative":
                self.abs_delaunay_threshold = self.delaunay_threshold * mean_min_dist
            elif self.delaunay_threshold_mode == "absolute":
                self.abs_delaunay_threshold = self.delaunay_threshold

        # Get the system dimensionality
        with chronic.Timer("TSA"):
            dimensionality = matid.geometry.get_dimensionality(
                system,
                self.cluster_threshold,
                dist_matrix_radii_pbc
            )
            if dimensionality is None:
                return Unknown(input_system)

        # 0D structures
        if dimensionality == 0:
            classification = Class0D(input_system)

            # Systems with one atom have their own classification.
            n_atoms = len(system)
            if n_atoms == 1:
                classification = Atom(input_system)

        # 1D structures
        elif dimensionality == 1:
            classification = Class1D(input_system)

        # 2D structures
        elif dimensionality == 2:

            classification = Class2D(input_system)

            # Get the indices of the used seed atoms
            seed_indices = []
            test_sys = system.copy()
            cm = matid.geometry.get_center_of_mass(test_sys)

            # If center of mass defined, for each atomic element find the
            # occurrence closest to center of mass to use as seed point.
            num = self.system.get_atomic_numbers()
            elems = set(num)

            if self.seed_position == "cm":
                distances = np.linalg.norm(system.get_positions() - cm, axis=1)
                indices = np.argsort(distances)
                for i in indices:
                    i_elem = num[i]
                    if i_elem in elems:
                        seed_indices.append(i)
                        elems.remove(i_elem)
                    if len(elems) == 0:
                        break
            else:
                if type(self.seed_position) == int:
                    seed_indices = [self.seed_position]
                elif isinstance(self.seed_position, (tuple, list, np.ndarray)):
                    seed_indices = self.seed_position

            # Find the best region by trying out different parameters options
            with chronic.Timer("cross_validation"):
                best_region = self.cross_validate_region(
                    system,
                    seed_indices,
                    disp_tensor_pbc,
                    disp_factors,
                    disp_tensor,
                    dist_matrix_radii_pbc
                )

            if best_region is not None:

                with chronic.Timer("region_analysis"):

                    # Check that the region was connected cyclically in two
                    # directions. This ensures that finite systems or systems
                    # with a dislocation at the cell boundary are filtered.
                    region_conn = best_region.get_connected_directions()
                    n_region_conn = np.sum(region_conn)
                    region_is_periodic = n_region_conn == 2
                    # cell_statistically_valid = best_region.get_cell_statistically_valid()
                    # print(cell_statistically_valid)

                    # This might be unnecessary because the connectivity of the
                    # unit cell is already checked.
                    clusters = best_region.get_clusters()
                    basis_indices = set(list(best_region.get_basis_indices()))
                    split = True
                    for cluster in clusters:
                        if basis_indices.issubset(cluster):
                            split = False

                    # Check that the found region covers enough of the entire
                    # system. If not, then the region alone cannot be used to
                    # classify the entire structure. This happens e.g. when one
                    # 2D sheet is found from a 2D heterostructure, or a local
                    # pattern is found inside a structure.
                    n_atoms = len(system)
                    n_basis_atoms = len(basis_indices)
                    coverage = n_basis_atoms/n_atoms
                    covered = coverage >= self.min_coverage

                    if not split and covered and region_is_periodic:
                        if best_region.is_2d:
                            classification = Material2D(input_system, best_region)
                        else:
                            classification = Surface(input_system, best_region)

        # Bulk structures
        elif dimensionality == 3:

            classification = Class3D(input_system)

            # Check the number of symmetries
            # analyzer = SymmetryAnalyzer(system)
            # crystallinity = matid.geometry.get_crystallinity(analyzer)
            # is_crystal = crystallinity >= self.crystallinity_threshold

            # If the structure is connected but the symmetry criteria was
            # not fullfilled, check the number of atoms in the primitive
            # cell. If above a certain threshold, try to find periodic
            # region to see if it is a crystal containing a defect.
            # if not is_crystal:
                # pass

                # This section is currently disabled. Can be reenabled once
                # more extensive testing is carried out on the detection of
                # defects in crystals.

                # primitive_system = analyzer.get_primitive_system()
                # n_atoms_prim = len(primitive_system)
                # if n_atoms_prim >= 20:
                    # periodicfinder = PeriodicFinder(
                        # pos_tol=self.abs_pos_tol,
                        # angle_tol=self.angle_tol,
                        # max_cell_size=self.max_cell_size,
                        # pos_tol_factor=self.pos_tol_factor,
                        # cell_size_tol=self.cell_size_tol,
                    # )

                    # # Get the index of the seed atom
                    # if self.seed_position == "cm":
                        # seed_vec = self.system.get_center_of_mass()
                    # else:
                        # seed_vec = self.seed_position
                    # seed_index = matid.geometry.get_nearest_atom(self.system, seed_vec)

                    # region = periodicfinder.get_region(system, seed_index, disp_tensor_pbc, disp_tensor, self.abs_delaunay_threshold)
                    # if region is not None:
                        # region = region[1]

                        # # If all the regions cover at least 80% of the structure,
                        # # then we consider it to be a defected crystal
                        # n_region_atoms = len(region.get_basis_indices())
                        # n_atoms = len(system)
                        # coverage = n_region_atoms/n_atoms
                        # if coverage >= self.coverage_threshold:
                            # classification = Crystal(analyzer, region=region)

            # elif is_crystal:
                # classification = Crystal(analyzer)

        return classification

    def cross_validate_region(
            self,
            system,
            seed_indices,
            disp_tensor_pbc,
            disp_factors,
            disp_tensor,
            dist_matrix_radii_pbc
        ):
        """Given a system tries multiple combinations of different search
        parameters to find a prototype cell and a corresponding region that
        best explains the underlying structure.

        Args:
        Returns:
        """
        # Run the detection with multiple position tolerances
        best_region = None
        most_atoms = 0

        # Here a cross-validation is performed to choose parameters that
        # produce best results. The performance of the parameters is
        # quantified by counting the number of valid atoms in the region.
        # The search is stopped if a system with zero outliers is found.
        n_atoms = len(system)
        for index in seed_indices:
            for size in self.max_cell_size:
                for tol in self.abs_pos_tol:

                    # Run the region detection on the whole system.
                    periodicfinder = PeriodicFinder(
                        angle_tol=self.angle_tol,
                        pos_tol_scaling=self.pos_tol_scaling,
                        cell_size_tol=self.cell_size_tol,
                        max_2d_cell_height=self.max_2d_cell_height,
                        max_2d_single_cell_size=self.max_2d_single_cell_size,
                        chem_similarity_threshold=self.chem_similarity_threshold
                    )
                    region = periodicfinder.get_region(
                        system,
                        index,
                        size,
                        tol,
                        self.abs_delaunay_threshold,
                        self.bond_threshold,
                        disp_tensor_pbc,
                        disp_factors,
                        disp_tensor,
                        dist_matrix_radii_pbc,
                    )

                    if region is not None:
                        basis_indices = region.get_basis_indices()
                        n_basis = len(basis_indices)

                        # There are no outliers with this cell, other
                        # options do not need to explored.
                        if n_basis == n_atoms:
                            return region

                        # Store this region if it is better than the
                        # previous best
                        if n_basis > most_atoms:
                            most_atoms = n_basis
                            best_region = region
        return best_region



from __future__ import absolute_import, division, print_function, unicode_literals

from ase import Atoms
import matid.geometry
import numpy as np


class System(Atoms):

    def __init__(
            self,
            symbols=None,
            positions=None,
            numbers=None,
            tags=None,
            momenta=None,
            masses=None,
            magmoms=None,
            charges=None,
            scaled_positions=None,
            cell=None,
            pbc=None,
            celldisp=None,
            constraint=None,
            calculator=None,
            info=None,
            wyckoff_letters=None,
            equivalent_atoms=None):

        super(System, self).__init__(
            symbols,
            positions,
            numbers,
            tags,
            momenta,
            masses,
            magmoms,
            charges,
            scaled_positions,
            cell,
            pbc,
            celldisp,
            constraint,
            calculator,
            info)

        self.wyckoff_letters = wyckoff_letters
        self.equivalent_atoms = equivalent_atoms

    @staticmethod
    def from_atoms(atoms):
        """Creates a System object from ASE.Atoms object.
        """
        system = System(
            positions=atoms.get_positions(),
            symbols=atoms.get_chemical_symbols(),
            cell=atoms.get_cell(),
            pbc=atoms.get_pbc(),
        )
        return system

    def to_scaled(self, positions, wrap=False):
        """Used to transform a set of positions to the basis defined by the
        cell of this system.

        Args:
            positions (numpy.ndarray): The positions to scale
            wrap (numpy.ndarray): Whether the positions should be wrapped
                inside the cell.

        Returns:
            numpy.ndarray: The scaled positions
        """
        return matid.geometry.to_scaled(
            self.get_cell(),
            positions,
            wrap,
            self.get_pbc())

    def to_cartesian(self, scaled_positions, wrap=False):
        """Used to transofrm a set of relative positions to the cartesian basis
        defined by the cell of this system.

        Args:
            positions (numpy.ndarray): The positions to scale
            wrap (numpy.ndarray): Whether the positions should be wrapped
                inside the cell.

        Returns:
            numpy.ndarray: The cartesian positions
        """
        return matid.geometry.to_cartesian(
            self.get_cell(),
            scaled_positions,
            wrap,
            self.get_pbc())

    def translate(self, translation, relative=False):
        """Translates the positions by the given translation.

        Args:
            translation (1x3 numpy.array): The translation to apply.
            relative (bool): True if given translation is relative to cell
                vectors.
        """
        matid.geometry.translate(self, translation, relative)

    def get_wyckoff_letters(self):
        """Returns a list of Wyckoff letters for the atoms in the system. This
        information is only available is explicitly set.

        Returns:
            np.ndarray: Wyckoff letters as a list of strings.
        """
        return np.array(self.wyckoff_letters)

    def set_wyckoff_letters(self, wyckoff_letters):
        """Used to set the Wyckoff letters of for the atoms in this system.

        Args:
            wyckoff_letters(sequence of str): The Wyckoff letters for the atoms
            in this system.
        """
        self.wyckoff_letters = np.array(wyckoff_letters)

    def get_equivalent_atoms(self):
        """Returns a list of indices marking the equivalence for the atoms in
        the system. This information is only available is explicitly set.

        Returns:
            np.ndarray: The equivalence information as a list of integers,
            where the same integer means equivalence and an integer is given
            for each atom.
        """
        return np.array(self.equivalent_atoms)

    def set_equivalent_atoms(self, equivalent_atoms):
        """Used to set the list of indices marking the equivalence for the
        atoms for the atoms in this system.

        Args:
            equivalent_atoms(sequence of int): list of indices marking the
                equivalence for the atoms the atoms in this system.
        """
        self.equivalent_atoms = np.array(equivalent_atoms)



import numpy as np


class WyckoffSet():
    """Represents a group of atoms in a certain Wyckoff position, for a certain
    space group.
    """
    def __init__(
            self,
            wyckoff_letter=None,
            atomic_number=None,
            element=None,
            indices=None,
            x=None,
            y=None,
            z=None,
            space_group=None,
            representative=None,
            multiplicity=None,
            ):
        """
        Args:
            wyckoff_letter (str): The letter for the Wyckoff position.
            atomic_number (int): Atomic number stored in this Wyckoff position.
            element (str): The chemical symbol of the element.
            indices (list): Indices corresponding to atoms in this Wyckoff
                group in a structure.
            x (float): The free parameter for the a-vector.
            y (float): The free parameter for the b-vector.
            z (float): The free parameter for the c-vector.
            space_group (int): The space group number for this set.
            representative (str): Algebraic expression that represents this set: all
                other positions are generated from symmetry.
            multiplicity (int): The multiplicity of the set: how many atoms are
                in this set.
        """
        self.wyckoff_letter = wyckoff_letter
        self.atomic_number = atomic_number
        self.element = element
        self.space_group = space_group
        self.indices = indices
        self.multiplicity = multiplicity
        self.representative = representative
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self): 
        msg = []
        msg.append("Wyckoff letter: {}, element: {}, multiplicity: {}, space group: {}".format(
            self.wyckoff_letter,
            self.element,
            self.multiplicity,
            self.space_group
        ))
        if self.x is not None:
            msg.append(", x: {}".format(self.x))
        if self.y is not None:
            msg.append(", y: {}".format(self.y))
        if self.z is not None:
            msg.append(", z: {}".format(self.z))
        return "".join(msg)
  
    def __str__(self): 
        return self.__repr__()

    def __eq__(self, other):
        if self.wyckoff_letter != other.wyckoff_letter: return False
        if self.atomic_number != other.atomic_number: return False
        if self.element != other.element: return False
        if self.space_group != other.space_group: return False
        if self.multiplicity != other.multiplicity: return False
        if self.x is not None and other.x is not None:
            if not np.isclose(self.x, other.x): return False
        else:
            if self.x != other.x: return False
        if self.y is not None and other.y is not None:
            if not np.isclose(self.y, other.y): return False
        else:
            if self.y != other.y: return False
        if self.z is not None and other.z is not None:
            if not np.isclose(self.z, other.z): return False
        else:
            if self.z != other.z: return False

        return True



from __future__ import absolute_import, division, print_function, unicode_literals

"""Used to store data related to atomic elements. Also has some conversion
utilities related to the data.
"""
import numpy as np
from ase.data import covalent_radii

SYMBOL_TO_NUMBER_MAP = {
    'H': 1,  'He': 2, 'Li': 3, 'Be': 4,
    'B': 5,  'C': 6,  'N': 7,  'O': 8,  'F': 9,
    'Ne': 10, 'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14,
    'P': 15,  'S': 16,  'Cl': 17, 'Ar': 18, 'K': 19,
    'Ca': 20, 'Sc': 21, 'Ti': 22, 'V': 23,  'Cr': 24,
    'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29,
    'Zn': 30, 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34,
    'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39,
    'Zr': 40, 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44,
    'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49,
    'Sn': 50, 'Sb': 51, 'Te': 52, 'I': 53,  'Xe': 54,
    'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59,
    'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64,
    'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69,
    'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74,
    'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79,
    'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84,
    'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89,
    'Th': 90, 'Pa': 91, 'U': 92,  'Np': 93, 'Pu': 94,
    'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99,
    'Fm': 100, 'Md': 101, 'No': 102, 'Lr': 103,
}
NUMBER_TO_SYMBOL_MAP = [
    None, 'H',  'He', 'Li', 'Be',
    'B',  'C',  'N',  'O',  'F',
    'Ne', 'Na', 'Mg', 'Al', 'Si',
    'P',  'S',  'Cl', 'Ar', 'K',
    'Ca', 'Sc', 'Ti', 'V',  'Cr',
    'Mn', 'Fe', 'Co', 'Ni', 'Cu',
    'Zn', 'Ga', 'Ge', 'As', 'Se',
    'Br', 'Kr', 'Rb', 'Sr', 'Y',
    'Zr', 'Nb', 'Mo', 'Tc', 'Ru',
    'Rh', 'Pd', 'Ag', 'Cd', 'In',
    'Sn', 'Sb', 'Te', 'I',  'Xe',
    'Cs', 'Ba', 'La', 'Ce', 'Pr',
    'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
    'Tb', 'Dy', 'Ho', 'Er', 'Tm',
    'Yb', 'Lu', 'Hf', 'Ta', 'W',
    'Re', 'Os', 'Ir', 'Pt', 'Au',
    'Hg', 'Tl', 'Pb', 'Bi', 'Po',
    'At', 'Rn', 'Fr', 'Ra', 'Ac',
    'Th', 'Pa', 'U',  'Np', 'Pu',
    'Am', 'Cm', 'Bk', 'Cf', 'Es',
    'Fm', 'Md', 'No', 'Lr'
]

"""List of atomic names in order."""
NUMBER_TO_NAME_MAP = [
    None,   'Hydrogen',  'Helium', 'Lithium', 'Beryllium',
    'Boron',  'Carbon',  'Nitrogen',  'Oxygen',  'Fluorine',
    'Neon', 'Sodium', 'Magnesium', 'Aluminum', 'Silicon',
    'Phosphorus',  'Sulfur',  'Chlorine', 'Argon', 'Potassium',
    'Calcium', 'Scandium', 'Titanium', 'Vanadium',  'Chromium',
    'Manganese', 'Iron', 'Cobalt', 'Nickel', 'Copper',
    'Zinc', 'Gallium', 'Germanium', 'Arsenic', 'Selenium',
    'Bromine', 'Krypton', 'Rubidium', 'Strontium', 'Yttrium',
    'Zirconium', 'Niobium', 'Molybdenum', 'Technetium', 'Ruthenium',
    'Rhodium', 'Palladium', 'Silver', 'Cadmium', 'Indium',
    'Tin', 'Antimony', 'Tellurium', 'Iodine',  'Xenon',
    'Cesium', 'Barium', 'Lanthanum', 'Cerium', 'Praseodymium',
    'Neodymium', 'Promethium', 'Samarium', 'Europium', 'Gadolinium',
    'Terbium', 'Dysprosium', 'Holmium', 'Erbium', 'Thulium',
    'Ytterbium', 'Lutetium', 'Hafnium', 'Tantalum', 'Tungsten',
    'Rhenium', 'Osmium', 'Iridium', 'Platinum', 'Gold',
    'Mercury', 'Thallium', 'Lead', 'Bismuth', 'Polonium',
    'Astatine', 'Radon', 'Francium', 'Radium', 'Actinium',
    'Thorium', 'Protactinium', 'Uranium',  'Neptunium', 'Plutonium',
    'Americium', 'Curium', 'Berkelium', 'Californium', 'Einsteinium',
    'Fermium', 'Mendelevium', 'Nobelium', 'Lawrencium'
]


def get_symbols(numbers):
    """Given atomic number(s), returns the symbol(s).

    Args:
        numbers (int or list of int): Atomic numbers(s) (number of protons).

    Returns:
        ndarray: Atomic symbol(s).

    Raises:
        ValueError: If a given atomic number is invalid and doesn't have a
        corresponding symbol.
    """
    error_msg = (
        "Given atomic number {} is invalid and doesn't have a symbol "
        "associated with it."
    )
    numbers = listify(numbers)
    symbols = []

    for number in numbers:
        try:
            symbol = NUMBER_TO_SYMBOL_MAP[number]
        except IndexError:
            raise ValueError(error_msg.format(number))
        if symbol is None:
            raise ValueError(error_msg.format(number))
        symbols.append(symbol)
    return symbols[0] if (len(numbers) == 1) else np.array(symbols, dtype=str)


def get_covalent_radii(species):
    """Give a list of species, returns the covalent radii.

    Args:
        species(list, str, int): One or multiple atomic species as symbols or
            atomic numbers.

    Returns:
        (numpy.ndarray): A numpy array of covalent radii.
    """

    numbers = listify(species)
    if isinstance(species[0], str):
        numbers = get_numbers(species)
    radii = []

    for number in numbers:
        radius = covalent_radii[number]
        radii.append(radius)
    return radii[0] if (len(radii) == 1) else np.array(radii, dtype=float)


def get_numbers(symbols):
    """Given element symbol(s), return the atomic number(s) (number of protons).

    Args:
        symbols (str or list of str): Atomic symbol(s).

    Returns:
        ndarray: Atomic number(s) (number of protons).

    Raises:
        ValueError: If a given atomic symbol is invalid and doesn't have a
        corresponding number.
    """
    symbols = listify(symbols)
    numbers = []

    for symbol in symbols:
        number = SYMBOL_TO_NUMBER_MAP.get(symbol)
        if number is None:
            raise ValueError(
                "Given atomic symbol {} is invalid and doesn't have a number "
                "associated with it.".format(symbol)
            )
        numbers.append(number)
    return numbers[0] if (len(number) == 1) else np.array(numbers, dtype=int)


def listify(value):
    """Given a value, force it into a list.

    Args:
        value:
    """
    is_string = isinstance(value, str)
    is_iterable = hasattr(value, '__iter__')

    if not is_string and is_iterable:
        return value
    else:
        return [value]



from __future__ import absolute_import, division, print_function, unicode_literals

from collections import defaultdict, OrderedDict
import numpy as np

import chronic

from ase import Atoms
from ase.data import covalent_radii
from ase.io import write

import matid.geometry
from matid.data import constants

import networkx as nx


class LinkedUnitCollection(dict):
    """Represents a collection of similar cells that are connected in 3D space
    to form a structure, e.g. a surface.

    Essentially this is a special flavor of a regular dictionary: the keys can
    only be a sequence of three integers, and the values should be LinkedUnits.
    """
    def __init__(
            self,
            system,
            cell,
            is_2d,
            dist_matrix_radii_pbc,
            disp_tensor_finite,
            delaunay_threshold=constants.DELAUNAY_THRESHOLD,
            chem_similarity_threshold=constants.CHEM_SIMILARITY_THRESHOLD,
            bond_threshold=constants.BOND_THRESHOLD,
        ):
        """
        Args:
            system(ase.Atoms): A reference to the system from which this
                LinkedUniCollection is gathered.
            cell(ase.Atoms): The prototype cell that is used in finding this
                region.
            is_2d(boolean): Whether this system represents a 2D-material or not.
            delaunay_threshold(float): The maximum allowed size of a tetrahedra
                edge in the Delaunay triangulation of the region..
        """
        self.system = system
        self.cell = cell
        self.is_2d = is_2d
        self.delaunay_threshold = delaunay_threshold
        self.chem_similarity_threshold = chem_similarity_threshold
        self.bond_threshold = bond_threshold
        self.dist_matrix_radii_pbc = dist_matrix_radii_pbc
        self.disp_tensor_finite = disp_tensor_finite
        self._search_graph = nx.MultiDiGraph()
        self._wrapped_moves = []
        self._index_cell_map = {}
        self._used_points = set()
        self._decomposition = None
        self._inside_indices = None
        self._outside_indices = None
        self._adsorbates = None
        self._substitutions = None
        self._vacancies = None
        self._clusters = None
        self._basis_indices = None
        self._basis_environments = None
        self._translations = None
        self._pos_tol = None
        dict.__init__(self)

    def __setitem__(self, key, value):
        # Transform key to tuple, check length
        try:
            key = tuple(key)
        except:
            raise TypeError(
                "Could not transform the given key '{}' into tuple."
                .format(key)
            )
        if len(key) != 3:
            raise ValueError(
                "The given coordinate '{}' does not have three components."
                .format(key)
            )

        # Check that old unit is not overwritten
        if key in dict.keys(self):
            raise ValueError(
                "Overriding existing units is not supported."
            )

        dict.__setitem__(self, key, value)

    def recreate_valid(self):
        """Used to recreate a new Atoms object, where each atom is created from
        a single unit cell. Atoms that were found not to belong to the periodic
        unit cell are not included.
        """
        recreated_system = Atoms(
            cell=self.system.get_cell(),
            pbc=self.system.get_pbc(),
        )
        for unit in self.values():
            i_valid_indices = np.array([x for x in unit.basis_indices if x is not None])
            if len(i_valid_indices) != 0:
                i_atoms = self.system[i_valid_indices]
                recreated_system += i_atoms

        return recreated_system

    def create_animation(self, ):
        """Used to create an animation of the search procedure. The unit cells are .
        """
        recreated_system = Atoms(
            cell=self.system.get_cell(),
            pbc=self.system.get_pbc(),
        )
        for iunit, unit in enumerate(self.values()):
            i_valid_indices = np.array([x for x in unit.basis_indices if x is not None])
            if len(i_valid_indices) != 0:
                i_atoms = self.system[i_valid_indices]
                recreated_system += i_atoms
                write('/home/lauri/Desktop/crystal/image_{}.png'.format(iunit), recreated_system, rotation='90x,20y,20x', show_unit_cell=2)

        # return recreated_system
        # for unit_cell in self.values():
            # rec = self.recreate_valid()
            # rec.set_cell(self.system.get_cell())
            # num = len(collection)
            # str_num = str(num)
            # str_len = len(str_num)
            # num = (3-str_len)*"0" + str_num
            # # write('/home/lauri/Desktop/2d/image_{}.png'.format(num), rec, rotation='-90x,45y,45x', show_unit_cell=2)
            # # write('/home/lauri/Desktop/2d/image_{}.png'.format(num), rec, rotation='', show_unit_cell=2)
            # # write('/home/lauri/Desktop/curved/image_{}.png'.format(num), rec, rotation='-80x', show_unit_cell=2)
            # # try:
            # write('/home/lauri/Desktop/crystal/image_{}.png'.format(num), rec, rotation='90x,20y,20x', show_unit_cell=2)
            # except Exception:
                # pass
            # else:
                # raise Exception("")

    def get_basis_atom_neighbourhood(self):
        """For each atom in the basis calculates the chemical neighbourhood.
        The chemical neighbourhood consists of a list of atomic numbers that
        are closer than a certain threshold value when the covalent radii is
        taken into account.

        Args:

        Returns:
        """
        if self._basis_environments is None:
            # Multiply the system to get the entire neighbourhood.
            cell = self.cell
            max_radii = covalent_radii[cell.get_atomic_numbers()].max()
            cutoff = max_radii + self.bond_threshold
            if self.is_2d:
                pbc = [True, True, False]
            else:
                pbc = [True, True, True]
            factors = matid.geometry.get_neighbour_cells(cell.get_cell(), cutoff, pbc)
            tvecs = np.dot(factors, cell.get_cell())

            # Find the factor corresponding to the original cell
            for i_factor, factor in enumerate(factors):
                if tuple(factor) == (0, 0, 0):
                    tvecs_reduced = np.delete(tvecs, i_factor, axis=0)
                    break

            pos = cell.get_positions()
            disp = matid.geometry.get_displacement_tensor(pos, pos, cell.get_cell())

            env_list = []
            for i in range(len(cell)):
                i_env = self.get_chemical_environment(
                    cell,
                    i,
                    disp,
                    tvecs,
                    tvecs_reduced
                )
                env_list.append(i_env)
            self._basis_environments = env_list

        return self._basis_environments

    def get_chem_env_translations(self):
        """Used to calculate the translations that are used in calculating the
        chemical enviroments.
        """
        if self._translations is None:
            cell = self.system.get_cell()
            num = self.system.get_atomic_numbers()
            max_radii = covalent_radii[num].max()
            cutoff = max_radii + self.bond_threshold
            factors = matid.geometry.get_neighbour_cells(cell, cutoff, True)
            translations = np.dot(factors, cell)

            # Find and remove the factor corresponding to the original cell
            for i_factor, factor in enumerate(factors):
                if tuple(factor) == (0, 0, 0):
                    translations_reduced = np.delete(translations, i_factor, axis=0)
                    break
            self._translations = (translations, translations_reduced)
        return self._translations

    def get_basis_indices(self):
        """Returns the indices of the atoms that were found to belong to a unit
        cell basis in the LinkedUnits in this collection as a single list.

        Returns:
            np.ndarray: Indices of the atoms in the original system that belong
            to this collection of LinkedUnits.
        """
        if self._basis_indices is None:
            with chronic.Timer("chemical_environment_translations"):
                translations, translations_reduced = self.get_chem_env_translations()

            # For each atom in the basis check the chemical environment
            with chronic.Timer("chemical_environment_basis"):
                neighbour_map = self.get_basis_atom_neighbourhood()

            with chronic.Timer("chemical_environment_units"):
                indices = set()
                for unit in self.values():

                    # Compare the chemical environment near this atom to the one
                    # that is present in the prototype cell. If these
                    # neighbourhoods are too different, then the atom is not
                    # counted as being a part of the region.
                    for i_index, index in enumerate(unit.basis_indices):
                        if index is not None:
                            real_environment = self.get_chemical_environment(self.system, index, self.disp_tensor_finite, translations, translations_reduced)
                            ideal_environment = neighbour_map[i_index]
                            chem_similarity = self.get_chemical_similarity(ideal_environment, real_environment)
                            if chem_similarity >= self.chem_similarity_threshold:
                                indices.add(index)

            # Ensure that all the basis atoms belong to the same cluster.
            # clusters = self.get_clusters()
            self._basis_indices = np.array(list(indices))

        return self._basis_indices

    def get_outliers(self):
        """Returns the indices of atoms that were not found to belong to the
        basis.
        """
        basis_indices = set(self.get_basis_indices())
        all_indices = set(self.get_all_indices())
        additional_indices = np.array(list(all_indices - basis_indices))

        return additional_indices

    def get_chemical_environment(
            self,
            system,
            index,
            disp_tensor_finite,
            translations,
            translations_reduced
        ):
        """Get the chemical environment around an atom. The chemical
        environment is quantified simply by the number of different species
        around a certain distance when the covalent radii have been considered.
        """
        # Multiply the system to get the entire neighbourhood.
        num = system.get_atomic_numbers()
        n_atoms = len(system)
        seed_num = num[index]

        neighbours = defaultdict(lambda: 0)
        for j in range(n_atoms):
            j_num = num[j]
            ij_disp = disp_tensor_finite[index, j, :]

            if index == j:
                trans = translations_reduced
            else:
                trans = translations

            D_trans = trans + ij_disp
            D_trans_len = np.linalg.norm(D_trans, axis=1)
            ij_radii = covalent_radii[seed_num] + covalent_radii[j_num]
            ij_n_neigh = np.sum(D_trans_len - ij_radii <= self.bond_threshold)

            neighbours[j_num] += ij_n_neigh

        return neighbours

    def get_chemical_similarity(self, ideal_env, real_env):
        """Returns a metric that quantifies the similarity between two chemical
        environments. Here the metric is defined simply by the number
        neighbours that are found to be same as in the ideal environmen within
        a certain radius.
        """
        max_score = sum(ideal_env.values())

        score = 0
        for ideal_key, ideal_value in ideal_env.items():
            real_value = real_env.get(ideal_key)
            if real_value is not None:
                score += min(real_value, ideal_value)

        return score/max_score

    def get_interstitials(self):
        """Get the indices of interstitial atoms in the original system.
        """
        inside_indices, _ = self.get_inside_and_outside_indices()
        inside_indices = set(inside_indices)
        substitutions = self.get_substitutions()
        subst_indices = set()
        for subst in substitutions:
            subst_indices.add(subst.index)
        interstitials = inside_indices - subst_indices

        return np.array(list(interstitials))

    def get_clusters(self):
        """Used to cluster the system in order to distinguish chemisorbed
        adsorbates from the surface.
        """
        if self._clusters is None:
            clusters = matid.geometry.get_clusters(
                self.dist_matrix_radii_pbc,
                self.bond_threshold
            )
            clusters = [set(list(x)) for x in clusters]
            self._clusters = clusters

        return self._clusters

    def get_adsorbates(self):
        """Get the indices of the adsorbate atoms in the region.

        All atoms that are outside the tesselation, and are either not part of
        the elements present in the surface or further away than a certain
        threshold, are labeled as adsorbate atoms.

        This function does not differentiate between different adsorbate
        molecules.

        Returns:
            np.ndarray: Indices of the adsorbates in the original system.
        """
        if self._adsorbates is None:

            _, outside_indices = self.get_inside_and_outside_indices()
            basis_elements = set(self.cell.get_atomic_numbers())
            outside_indices = outside_indices

            # The substitutions have to be removed explicitly from the ouside
            # atoms because sometimes they are outside the tesselation.
            substitutions = self.get_substitutions()
            outside_indices = set(outside_indices)
            substitutions = set([x.index for x in substitutions])
            outside_indices -= substitutions
            outside_indices = np.array(list(outside_indices))

            if len(outside_indices) != 0:

                basis_elements = set(basis_elements)
                adsorbates = []
                for index in outside_indices:
                    adsorbates.append(index)
                adsorbates = np.array(adsorbates)
            else:
                adsorbates = np.array([])

            self._adsorbates = adsorbates

        return self._adsorbates

    def get_substitutions(self):
        """Get the substitutions in the region.
        """
        if self._substitutions is None:

            # Gather all substitutions
            # all_substitutions = []
            # for cell in self.values():
                # subst = cell.substitutions
                # if len(subst) != 0:
                    # all_substitutions.extend(subst)

            # The substitutions are validate based on their chemical
            # environment and position in the triangulation.
            neighbour_map = self.get_basis_atom_neighbourhood()
            valid_subst = []
            # _, outside_indices = self.get_inside_and_outside_indices()
            # outside_set = set(outside_indices)
            translations, translations_reduced = self.get_chem_env_translations()

            for unit in self.values():

                # Compare the chemical environment near this atom to the one
                # that is present in the prototype cell. If these
                # neighbourhoods are too different, then the atom is not
                # counted as being a part of the region.
                if unit.substitutions is not None:
                    for i_index, subst in enumerate(unit.substitutions):
                        if subst is not None:
                            subst_index = subst.index

                            # Otherwise check the chemical similarity
                            real_environment = self.get_chemical_environment(
                                self.system,
                                subst_index,
                                self.disp_tensor_finite,
                                translations,
                                translations_reduced
                            )
                            ideal_environment = neighbour_map[i_index]
                            chem_similarity = self.get_chemical_similarity(ideal_environment, real_environment)
                            if chem_similarity >= self.chem_similarity_threshold:
                                valid_subst.append(subst)
            self._substitutions = valid_subst

            # In 2D materials all substitutions in the cell are valid
            # substitutions
            # if self.is_2d:
                # self._substitutions = all_substitutions
            # else:
                # # In surfaces the substitutions have to be validate by whether they
                # # are inside the tesselation or not
                # inside_indices, _ = self.get_inside_and_outside_indices()
                # inside_set = set(inside_indices)

                # # Find substitutions that are inside the tesselation
                # valid_subst = []
                # for subst in all_substitutions:
                    # subst_index = subst.index
                    # if subst_index in inside_set:
                        # valid_subst.append(subst)
                # self._substitutions = valid_subst

        return self._substitutions

    def get_vacancies(self):
        """Get the vacancies in the region.

        Returns:
            ASE.Atoms: An atoms object representing the atoms that are missing.
            The Atoms object has the same properties as the original system.
        """
        if self._vacancies is None:

            # Gather all vacancies
            all_vacancies = []
            for cell in self.values():
                vacancies = cell.vacancies
                if len(vacancies) != 0:
                    all_vacancies.extend(vacancies)

            # For purely 2D systems all missing atoms in the basis are vacancies
            if self.is_2d:
                self._vacancies = all_vacancies
            else:
                # Get the tesselation
                tesselation = self.get_tetrahedra_decomposition()

                # Find substitutions that are inside the tesselation
                valid_vacancies = []
                for vacancy in all_vacancies:
                    vac_pos = vacancy.position
                    simplex = tesselation.find_simplex(vac_pos)
                    if simplex is not None:
                        valid_vacancies.append(vacancy)
                self._vacancies = valid_vacancies

        return self._vacancies

    def get_tetrahedra_decomposition(self):
        """Get the tetrahedra decomposition for this region.
        """
        if self._decomposition is None:
            # Get the positions of basis atoms
            basis_indices = self.get_basis_indices()
            valid_sys = self.system[basis_indices]

            # Perform tetrahedra decomposition
            self._decomposition = matid.geometry.get_tetrahedra_decomposition(
                valid_sys,
                self.delaunay_threshold
            )

        return self._decomposition

    def get_all_indices(self):
        """Get all the indices that are present in the full system.
        """
        return set(range(len(self.system)))

    def get_unknowns(self):
        """Returns indices of the atoms that are in the outliers but are not
        recognized as any specialized group.
        """
        outliers = set(self.get_outliers())
        adsorbates = set(self.get_adsorbates())
        interstitials = set(self.get_interstitials())
        substitutions = set([x.index for x in self.get_substitutions()])

        return outliers - adsorbates - interstitials - substitutions

    def get_inside_and_outside_indices(self):
        """Get the indices of atoms that are inside and outside the tetrahedra
        tesselation.

        Returns:
            (np.ndarray, np.ndarray): Indices of atoms that are inside and
            outside the tesselation. The inside indices are in the first array.
        """
        if self._inside_indices is None and self._outside_indices is None:
            invalid_indices = self.get_outliers()
            invalid_pos = self.system
            inside_indices = []
            outside_indices = []

            if len(invalid_indices) != 0:
                invalid_pos = self.system.get_positions()[invalid_indices]
                tesselation = self.get_tetrahedra_decomposition()
                for i, pos in zip(invalid_indices, invalid_pos):
                    simplex_index = tesselation.find_simplex(pos)
                    if simplex_index is None:
                        outside_indices.append(i)
                    else:
                        inside_indices.append(i)

            self._inside_indices = np.array(inside_indices)
            self._outside_indices = np.array(outside_indices)

        return self._inside_indices, self._outside_indices

    def get_connected_directions(self):
        """During the tracking of the region the information about searches
        that matched an atom twive but with a negated multiplier are stored.
        """
        connected_directions = np.array([False, False, False])

        # Find all the nodes that have at least two incoming edges. If there
        # are two incoming edges with negated multiplier, there is periodicity
        # in the multiplier direction.
        G = self._search_graph
        for node in G.nodes():
            node_edges = G.in_edges(node, data=True)
            multiplier_sum = np.array([0, 0, 0])
            multiplier_presence = np.array([False, False, False])
            for edge in node_edges:
                source = edge[0]
                dest = edge[1]
                unit_index_change = np.array(dest) - np.array(source)
                multiplier = edge[2]["multiplier"]
                multiplier_sum += multiplier

                # Create a mask that selects the index where the move occurred
                move_mask = multiplier != 0

                # If the movement does not correspond to the change in unit
                # cell index, then the movement has wrapped across and that
                # direction is periodic
                if multiplier[move_mask] != unit_index_change[move_mask]:
                    multiplier_presence[move_mask] = True

            for i in range(3):
                if multiplier_presence[i]:
                    if multiplier_sum[i] == 0:
                        connected_directions[i] = True

        # For each loop, we calculate the total displacement. If it is nonzero,
        # the nonzero direction is marked as a connected direction.
        # cycles = list(nx.simple_cycles(self._search_graph))
        # loop_multipliers = []
        # for i_loop, loop in enumerate(cycles):
            # loop_len = len(loop)

            # # A self-loop can be formed when the found vector corresponds to a
            # # simulation vector. In this case the loop has only one element,
            # # and all the edges are valid multipliers.
            # if loop_len == 1:
                # source = loop[0]
                # dest = loop[0]
                # edges = G[source][dest]
                # for key, edge in edges.items():
                    # multiplier = edge["multiplier"]
                    # loop_multipliers.append(multiplier)
            # # A loop between two atoms can be formed when there are only two
            # # repetitions of the cell per simulation cell basis vector. In any
            # # of the preset multipliers will be valid. Two-element loops within
            # # the simulation cell cannot be formed because the search never
            # # come back to an atom without crossing the periodic boundary.
            # elif loop_len == 2:
                # source = loop[0]
                # dest = loop[1]
                # edges = G[source][dest]
                # for key, edge in edges.items():
                    # multiplier = edge["multiplier"]
                    # if multiplier[0] != 0:
                        # print(multiplier)
                        # print(G.nodes[source])
                        # print(G.nodes[dest])
                    # loop_multipliers.append(multiplier)
            # # We can ignore loops with more elements. The two-body loops will
            # # already tell the directions in which the graph is periodic.
            # else:
                # pass

        # loop_multipliers = np.array(loop_multipliers)
        # indices = np.where(loop_multipliers != 0)[1]
        # indices = np.unique(indices)
        # connected_directions[indices] = True

        return connected_directions

    # def get_cell_statistically_valid(self):
        # """Checks that in this region a certain fraction of the cells is
        # complete, i.e. has all the atoms that were found for the prototype
        # cell.
        # """
        # n_all = 0
        # index_occurrence_map = defaultdict(lambda: 0)
        # for cell in self.values():
            # none_found = False
            # index_found = False
            # for i, basis_index in enumerate(cell.basis_indices):
                # if basis_index is not None:
                    # index_occurrence_map[i] += 1
                    # index_found = True
                # else:
                    # none_found = True
            # if none_found:
                # if index_found:
                    # n_all += 1

        # for index, value in index_occurrence_map.items():
            # if value/n_all <= 0.5:
                # return False
        # return True

        # full_fraction = n_full / n_all
        # return full_fraction > 0.5


class LinkedUnit():
    """Represents a cell that is connected to others in 3D space to form a
    structure, e.g. a surface.
    """
    def __init__(self, index, seed_index, seed_coordinate, cell, basis_indices, substitutions, vacancies):
        """
        Args:
            index(tuple of three ints):
            seed_index(int):
            seed_coordinate():
            cell(np.ndarray): Cell for this unit. Can change from unit to unit.
            basis_indices(sequence of ints and Nones): A sequence where there
                is an index or None for each atom that is supposed to be in the
                basis
            substitute_indices(sequence of ints and Nones): If basis atom is
                replaced by a foreign atom, the index of the substitutional atom is
                here.
        """
        self.index = index
        self.seed_index = seed_index
        self.seed_coordinate = seed_coordinate
        self.cell = cell
        self.basis_indices = basis_indices
        self.substitutions = substitutions
        self.vacancies = vacancies


class Substitution():
    """Represents a substitutional point defect.
    """
    def __init__(self, index, position, original_element, substitutional_element):
        """
        Args:
            index (int): Index of the subtitutional defect in the original
                system.
            original (ase.Atom): The original atom that was substituted
            substitution (ase.Atom): The substitutional atom

        """
        self.index = index
        self.original_element = original_element
        self.substitutional_element = substitutional_element
